# (436)
# (190)
# eg : HIVE-494fc462-e060-4617-9baf-3b35793a37f8
"""
Strategie

494fc462-e060-4617-9baf-3b35793a37f8

1. Retrieve callerId form  `submitDAGSession(429)`:

2020-09-15 22:34:41,833 INFO  [HiveServer2-Background-Pool: Thread-199227]: client.TezClient (TezClient.java:submitDAGSession(429)) - Submitting dag to TezSession, sessionName=HIVE-ec1e078a-5e58-47c6-aa33-cdd221b3f5b3, applicationId=application_1598886766592_49581, dagName=SELECT * FROM rep_prd...sc_quote_line_item_c(Stage-1), callerContext={ context=HIVE, callerType=HIVE_QUERY_ID, callerId=hive_20200915175336_89c35e6d-0b15-4f26-bba0-67f0185e3a36 }

2. With callerid you can retrieve `execute(1423)` which contains the query

2020-09-15 14:40:52,517 INFO  [HiveServer2-Background-Pool: Thread-182868]: ql.Driver (Driver.java:execute(1423)) - Starting command(queryId=hive_20200915144052_e80d53b2-832e-42a9-88aa-4fe3e5c07bab): SELECT num_service FROM rep_prd_fbo_project.parametres_cedre_c

eg. grep like:

➜ zgrep -E 'TezClient.java:submitDAGSession|Driver.java:execute' hiveserver2.log > query.log

"""
from typing import List, Pattern, Iterator
from pathlib import Path
from dataclasses import dataclass, field
import gzip
import re
import pandas as pd


def grep_multiline_gz(path: Path, token: str, newline: Pattern) -> List[str]:
    logfile = gzip.open(path, 'rt')
    return grep_multiline(logfile, token, newline)


def grep_multiline(lines: Iterator, token: str, newline: Pattern) -> List[str]:
    """
    NOTE: it will skip a line don't use it as a standard solution
    """
    result = []
    for line in lines:
        chunk = []
        if token in line:
            chunk.append(line)
            for subline in lines:
                if newline.match(subline):
                    result.append(''.join(chunk))
                    break
                else:
                    # we'll skip a line in main for loop
                    # but we don't mind as we expect each relevant line
                    # to have several lines between them
                    chunk.append(subline)
    return result


@dataclass
class HiveServer:
    logs: List[str]
    newline: Pattern = re.compile(r'(\d+-?){3}\s(\d+[:,]?){4}')
    # TezClient.java:submitDAGSession(429)
    submit_dag: Pattern = re.compile(
        r'(?P<ts>(\d+-?){3}\s(\d+[:,]?){4})\s(?P<level>[A-Z]+)\s{2}.*Thread-(?P<thread>[0-9]+)(.*)(?P<id_application>application_[0-9]+_[0-9]+)', re.S)
    # Driver.java.execute
    driver_execute: Pattern = re.compile(
        r'(?P<ts>(\d+-?){3}\s(\d+[:,]?){4})\s(?P<level>[A-Z]+)\s{2}.*Thread-(?P<thread>[0-9]+).*(?P<id_caller>hive[0-9_\-a-z]+)\):\s(?P<query>.*)', flags=re.S)
    mapping: List[dict] = field(default_factory=list)
    queries: List[dict] = field(default_factory=list)
    token_mapping = 'TezClient.java:submitDAGSession(429)'
    token_query = 'Driver.java:execute(1423)'
    dataset: pd.DataFrame = pd.DataFrame()

    def __post_init__(self):
        self.mapping = [self.parse_submit_dag(
            line) for line in self.logs if self._is_mapping(line)]
        self.queries = [self.parse_driver_execute(
            line) for line in self.logs if self._is_query(line)]
        if len(self.mapping) > 0:
            df_mapping = pd.DataFrame(self.mapping).drop(
                columns=['ts', 'level'])
            df_queries = pd.DataFrame(self.queries)
            self.dataset = pd.merge(
                df_mapping, df_queries, on='thread').set_index('id_application')

    @classmethod
    def from_file(cls, log: Path):
        cls.token_mapping
        lines_mapping = grep_multiline(
            log.open('r'), cls.token_mapping, cls.newline)
        lines_query = grep_multiline(
            log.open('r'), cls.token_query, cls.newline)
        return cls(lines_mapping+lines_query)

    @classmethod
    def from_gz_file(cls, log: Path):
        cls.token_mapping
        lines_mapping = grep_multiline(
            gzip.open(log, 'rt'), cls.token_mapping, cls.newline)
        lines_query = grep_multiline(
            gzip.open(log, 'rt'), cls.token_query, cls.newline)
        return cls(lines_mapping+lines_query)

    def _is_mapping(self, line: str) -> bool:
        return self.token_mapping in line

    def _is_query(self, line: str) -> bool:
        return self.token_query in line

    @staticmethod
    def parse_line(line: str, pattern: Pattern) -> dict:
        try:
            return pattern.match(line).groupdict()
        except AttributeError:
            return {}

    def parse_submit_dag(self, line) -> dict:
        return self.parse_line(line, self.submit_dag)

    def parse_driver_execute(self, line) -> dict:
        return self.parse_line(line, self.driver_execute)

    def get_thread(self, app_id: str) -> str:
        try:
            return self.dataset.loc[app_id, 'thread']
        except KeyError:
            return ''

    def get_query(self, app_id: str) -> str:
        try:
            return self.dataset.loc[app_id, 'query']
        except KeyError:
            return ''



from typing import List
import humanize
from datetime import timedelta, datetime
import pandas as pd


def human_elapsed(nb: int) -> str:
    """milliseconds to human readable"""
    _t = humanize.i18n.activate("fr")
    return humanize.precisedelta(timedelta(milliseconds=nb), minimum_unit='hours')


def human_dt(dt: datetime) -> str:
    return dt.strftime('%d/%m/%Y à %Hh%M')

def flatten(raw: dict) -> list:
    return [elt for elt in raw['apps']['app']]

def to_minutes(milliseconds:int)->int:
    return round((milliseconds/1000)/60)


def get_long_run(payload: dict, threshold: int,queue = '') -> List[dict]:
    """
    Retrieve all application currently running for more than `threshold` minutes

    Parameters
    ----------
    payload:
        json returned py Yarn TimeServer
    threshold:
        number of minutes
    
    """
    if not payload:
        return []
    apps = flatten(payload)
    filtered = [elt for elt in apps if elt['state'] == 'RUNNING' and queue in elt['queue'].lower()]
    if not filtered:
        return []
    df = pd.DataFrame(filtered)
    for dt in ['startedTime', 'finishedTime']:
        df[dt] =pd.to_datetime(df[dt],unit='ms')
    df['elapsed_minutes'] = df['elapsedTime'].apply(to_minutes)
    df['started_time'] = df.startedTime.dt.strftime('%d/%m/%Y %Hh%M')
    df_threshold = df[df.elapsed_minutes>threshold]
    
    return df_threshold[['id','user','started_time','queue','elapsed_minutes']].to_dict(orient='records')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['khadoop', 'khadoop.hiveserver', 'khadoop.yarn']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=0.15.8,<0.16.0', 'humanize>=3.1.0,<4.0.0', 'pandas>=1.1,<2.0']

setup_kwargs = {
    'name': 'khadoop',
    'version': '1.3.3',
    'description': '',
    'long_description': "# README\n\nParse and slice hadoop logs\n\n## Yarn RM\n\n![alt](img/yarn-rm.png)\n\n### Dataset\n\n```python\nfrom khadoop.yarn import logrm\n```\n\nParse all files that look like a regular Ressource Manager log with default name.\n\n`logrm.FILEPATTERN` is a unix-like pattern file to help glob them.\n\n```python\nparsed = []\nfor filelog in LOGFOLDER.glob(logrm.FILEPATTERN):\n    print(filelog)\n    parsed += logrm.process(filelog.open())\n```\n\n`logrm.process` will parse each line and produce a list of dict with sensible information\n\neach dict look like :\n\n```python\n {\n   'accepted_to_running': 6,  # nb sec between ACCEPT to RUNNING\n   'id_application': 'application_1596547077642_6854',\n   'accept_to_running_ts':'2020-08-06 14:59:59,119' # timestamp set for log line 'FROM accepted to RUNNING'\n   }\n```\n\nthe `accepted_to_running` represent here the number between these two timestamps on yarn aggregated RM log:\n\n```log\n2020-08-06 14:59:52,756 INFO  rmapp.RMAppImpl (RMAppImpl.java:handle(779)) - application_1596547077642_6854 State change from SUBMITTED to ACCEPTED\n...\n2020-08-06 14:59:59,119 INFO  rmapp.RMAppImpl (RMAppImpl.java:handle(779)) - application_1596547077642_6854 State change from ACCEPTED to RUNNING\n```\n\n## Related\n\n- https://github.com/etsy/logster\n",
    'author': 'Khalid',
    'author_email': 'khalidck@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
